# encoding: utf-8
import logging

import re

import os
import socket

import sys
import threading

import time
from urllib import request
from . import boot


def read_self_prop():
    try:
        def_prop = __import__('dophon.def_prop.default_properties', fromlist=True)
        u_prop = __import__('application', fromlist=True)
        # 对比配置文件
        for name in dir(def_prop):
            if re.match('__.*__', name):
                continue
            if name in dir(u_prop):
                continue
            setattr(u_prop, name, getattr(def_prop, name))
        sys.modules['properties'] = u_prop
        sys.modules['dophon.properties'] = u_prop
    except Exception as e:
        logging.error(e)
        sys.modules['properties'] = def_prop
        sys.modules['dophon.properties'] = def_prop


try:
    read_self_prop()
except Exception as e:
    logging.error('没有找到自定义配置:(application.py)')
    logging.error('引用默认配置')

from dophon import properties
from dophon_logger import *

logger = get_logger(DOPHON)

logger.inject_logger(globals())


def IsOpen(ip, port):
    """
    检查端口是否被占用
    :param ip:
    :param port:
    :return:
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        s.connect((ip, int(port)))
        logger.error('端口被占用:' + port)
        s.close()
        return True
    except:
        return False


def listen(code):
    """
    监听命令状态码
    :param code: os.system()命令返回码
    :return: 当不正常执行会抛出错误,谨慎使用!!!
    """
    if code >> 8 is 0:
        pass
    else:
        raise Exception('命令执行错误!')


def listen_container_status(container_port, loop_count: int = 3, wait_sec: int = 10):
    """
    检测容器端口存活
    :return:
    """
    # 一分钟后检测
    time.sleep(60)
    # 默认检测三次
    curr_count = 1
    while int(curr_count) <= int(loop_count):
        # 默认间隔10秒
        time.sleep(wait_sec)
        if IsOpen(get_docker_address(), int(container_port)):
            raise Exception('端口映射异常')
        else:
            # 报错证明端口正常占用
            # 发起请求
            url = 'http://' + get_docker_address() + ':' + container_port
            logger.info('容器存活性检查:' + url)
            res = request.urlopen(url)
            if not res.read():
                raise Exception('服务启动异常')
        curr_count += 1
    logger.info('容器启动成功,请在命令行输入docker ps查看')


def get_docker_address():
    """
    获取容器载体ip
    :return:
    """
    result = os.popen('ipconfig').readlines()
    result_lines = []
    r_l_copy = []
    while result:
        line = result[0]
        if re.search('^.*(d|D)(o|O)(c|C)(k|K)(e|E)(r|R).*$', line):
            result_lines = result.copy()
            break
        else:
            result.pop(0)
    for line in result_lines:
        line = re.sub('\s*', '', line)
        if line and re.search('([0-9]+\.)+[0-9]+$', line) and re.search('(i|I)(p|P)', line):
            r_l_copy.append(
                re.search('([0-9]+\.[0-9]+\.[0-9]+\.[0-9]+)', line).group(0)
            )
    return r_l_copy.pop(0) if r_l_copy else 'can not get container ip'


def attach_container(base_name: str):
    """
    进入容器
    :return:
    """
    os.system('docker attach ' + base_name)


def run_as_docker(
        entity_file_name: str = None,
        container_port: str = str(properties.port),
        docker_port: str = str(properties.port),
        attach_cmd: bool = False,
        extra_package: dict = {},
        cache_virtual_env_dir: str = '',
        package_repository: str = '',
):
    """
    利用docker启动项目
    :param entity_file_name: 入口文件名(包括后缀)
    :param container_port: 容器暴露端口
    :param docker_port: 容器内部端口 -> 集群模式下的暴露端口,一般为配置文件定义的端口
    :param attach_cmd: 是否进入容器内部sh
    :param extra_package: 额外需要加载的包以及版本
    :param cache_virtual_env_dir: 指定的虚拟器路径
    :param package_repository: 自带缓存包路径,若为空会自动执行pip安装
                # 阿里云仓库  =>  https://mirrors.aliyun.com/pypi/simple/
    :return:
    """
    try:
        logger.info('容器前期准备')
        root = re.sub('\\\\', '/', properties.project_root)
        base_name = os.path.basename(root)
        import platform
        p_version = platform.python_version()
        work_dir = './' + base_name
        # 生成依赖文件
        logger.info('生成依赖文件')
        os.system('pip freeze --all >pre_requirements.txt')
        with open('./pre_requirements.txt', 'r') as file:
            with open('./requirements.txt', 'w') as final_file:

                for line in file.readlines():
                    for key in sys.modules.keys():
                        if re.search('(_|__|\\.).+$', key):
                            continue
                        module_path = re.sub('(>=|==|=>|<=|=<|<|>|=).*\s+', '', line.lower())
                        if re.search(module_path, key.lower()):
                            if module_path in extra_package:
                                final_file.write(f'{module_path}>={extra_package[module_path]}\n')
                                extra_package.pop(module_path)
                            else:
                                final_file.write(line)

                            continue
                # 写入额外包
                for package_name, package_version in extra_package.items():
                    if package_name in extra_package:
                        final_file.write(''.join([package_name, '>=', extra_package[package_name], '\n']))
                        extra_package.popitem()
                    else:
                        final_file.write(''.join([package_name, '>=', package_version, '\n']))
        # 生成Dockerfile
        logger.info('生成Dockerfile')
        with open('./Dockerfile', 'w') as file:
            file.write('FROM python:' + p_version + '\n')
            file.write('ADD . ' + work_dir + '\n')
            file.write('ADD . ' + work_dir + '/' + base_name + '\n')
            file.write('WORKDIR ' + work_dir + '\n')
            if cache_virtual_env_dir:
                file.write(f'ADD {cache_virtual_env_dir} ~/.cache_virtual_env' + '\n')
                file.write(f'CMD ~/.cache_virtual_env/Scripts/activate' + '\n')
            if package_repository:
                # 阿里云仓库  =>  https://mirrors.aliyun.com/pypi/simple/
                file.write(f'RUN pip install -i {package_repository} -r requirements.txt' + '\n')
            else:
                file.write('RUN pip install --no-cache-dir -r requirements.txt' + '\n')
            file.write('CMD ["python","./' + (entity_file_name if entity_file_name else 'Bootstrap.py') + '"]' + '\n')
            # file.write('CMD ["/bin/bash"]' + '\n')
        os.system('cd ' + root)
        logger.info('暂停已运行的实例')
        os.system('docker stop ' + base_name)
        logger.info('移除已运行的实例')
        os.system('docker rm ' + base_name)
        logger.info('移除旧镜像')
        os.system('docker rmi ' + base_name)
        logger.info('检测配置合法性')
        if IsOpen('127.0.0.1', int(docker_port)):
            # 端口被占用
            logger.warn('映射端口被占用!!')
        logger.info('建立镜像')
        listen(os.system('docker build -t ' + base_name + ' .'))
        logger.info('运行镜像')
        os.system(
            'docker run -p ' + container_port
            +
            ':' +
            docker_port +
            ' -d --name ' +
            base_name + ' ' +
            os.path.basename(
                root))
        logger.info('打印容器内部地址(空地址代表启动失败)')
        os.system('docker inspect --format=\'{{.NetworkSettings.IPAddress}}\' ' + base_name)
        logger.info('打印容器载体地址')
        print(get_docker_address())
        logger.info('启动检测容器端口')
        threading.Thread(target=listen_container_status, args=(container_port,)).start()
        if attach_cmd:
            logger.info('进入镜像')
            # threading.Thread(target=attach_container,args=(base_name,)).start()
            attach_container(base_name)
        logger.info('容器启动完毕')
    except Exception as e:
        logger.error(e)



import gc

gc.enable()  # 开启gc


def show_gc_leak(method):
    """
    打印并获取gc信息
    1.不可达对象统计
    2.不可回收对象统计
    :param method: 打印函数引用
    :return:
    """
    _unreachable = gc.collect()
    method(
        '%s, %s, %s, %s, %s' %
        (
            "unreachable object num: %d" % (_unreachable,),
            "garbage object num: %d" % (len(gc.garbage)),
            "gc object num: %d" % (len(gc.get_objects())),
            "gc threshold num: %d" % (len(gc.get_threshold())),
            "gc stats num: %d" % (len(gc.get_stats())),
        )
    )
    return {
        'unreachable object num': _unreachable,
        'garbage object num': len(gc.garbage),
        "gc object num": len(gc.get_objects()),
        # "gc object info": gc.get_objects(),
        "gc threshold num": len(gc.get_threshold()),
        "gc threshold info": str(gc.get_threshold()),
        "gc stats num": len(gc.get_stats()),
        "gc stats info": str(gc.get_stats()),
    }



# 响应返回数据修饰器(跳过视图解析器)
# 响应体形式返回
# (自带AutoParam) <------暂未完成
import functools
from dophon_logger import *
from flask import jsonify, render_template

logger = get_logger(DOPHON)

logger.inject_logger(globals())

def response_body():
    def method(f):
        @functools.wraps(f)
        def arg(*args, **kwargs):
            result = jsonify(f(*args, **kwargs))
            if result:
                return result
            else:
                return []

        return arg

    return method


# 返回web模板
def response_template(template: list):
    """
    返回模板页面
    :param template: 模板页面路径
    :return:
    """

    def method(f):
        @functools.wraps(f)
        def args(*args, **kwargs):
            page_param = f(*args, **kwargs)
            if isinstance(page_param, type({})):
                return render_template(template, **page_param)
            else:
                logger.error('页面参数错误!')
                raise KeyError('页面参数错误!')

        return args

    return method


# coding: utf-8
import functools
import re
from flask import request, abort
from dophon_logger import *

logger = get_logger(DOPHON)

'''
参数体可以为多个,形参名称必须与请求参数名一一对应(只少不多)
装饰器中关键字参数列表可以指定参数名
'''

logger.inject_logger(globals())


# 处理请求参数装饰器(分离参数)


def auto_param(kwarg_list=[]):
    def method(f):
        @functools.wraps(f)
        def args(*args, **kwargs):
            try:
                if 'GET' == str(request.method):
                    return is_get(f, request, kwarg_list)
                elif 'POST' == str(request.method):
                    return is_post(f, request, kwarg_list)
                else:
                    logger.error('方法不支持!!')
            except TypeError as t_e:
                logger.error('参数不匹配!!,msg:' + repr(t_e))
                raise t_e
                return abort(500)

        return args

    return method


'''
注意!!!
该注解只需方法体内存在一个形参!!!
同样,需要指定参数名的形式参数列表条目也只能存在一个,多个会默认取第一个
不匹配会打印异常
参数以json形式赋值
'''


# 处理请求参数装饰器(统一参数,参数体内参数指向json串)
def full_param(kwarg_list=[]):
    def method(f):
        @functools.wraps(f)
        def args(*args, **kwargs):
            try:
                if 'POST' == str(request.method):
                    r_arg = ()
                    r_kwarg = {}
                    if not kwarg_list:
                        r_arg = (request.json if request.is_json else request.form.to_dict(),)
                    else:
                        r_kwarg[kwarg_list[0]] = request.json if request.is_json else request.form.to_dict()
                    return f(*r_arg, **r_kwarg)
                elif 'GET' == str(request.method):
                    r_arg = ()
                    r_kwarg = {}
                    if not kwarg_list:
                        r_arg = (request.args.to_dict(),)
                    else:
                        r_kwarg[kwarg_list[0]] = request.args.to_dict()
                    return f(*r_arg, **r_kwarg)
                else:
                    logger.error('json统一参数不支持该请求方法!!')
                    return abort(400)
            except TypeError as t_e:
                logger.error('参数不匹配!!,msg:' + repr(t_e))
                return abort(500)

        return args

    return method


def file_param(alias_name: str = 'files', extra_param: str = 'args'):
    """
    文件参数装在装饰器

    ps 文件上传暂时只支持路由方法内单个参数接收(会有校验策略)

    参数demo(小程序):
        ImmutableMultiDict([('img_upload_test', <FileStorage: 'filename' ('image/jpeg')>)])
    :return:
    """

    def method(f):
        @functools.wraps(f)
        def args(*args, **kwargs):
            # 检测参数
            a_nums = len(args) + len(kwargs)
            if a_nums > 0:
                logger.error('路由绑定参数数量异常')
                raise Exception('路由绑定参数数量异常')
            try:
                extra_param_value = (request.form if request.form else request.json).to_dict()
            except:
                extra_param_value = {}
            k_args = {
                alias_name: request.files.to_dict(),
                extra_param: extra_param_value
            }
            return f(**k_args)

        return args

    return method


def is_get(f, r, kw_list):
    r_arg = ()
    r_kwarg = {}
    if not kw_list:
        r_arg = r.args.to_dict().values()
    else:
        for index in range(len(kw_list)):
            r_kwarg[kw_list[index]] = r.args[kw_list[index]]
    return f(*r_arg, **r_kwarg)


def is_post(f, r, kw_list):
    r_arg = ()
    r_kwarg = {}
    if r.is_json:
        if not kw_list:
            r_arg = is_json(r.json)
        else:
            for index in range(len(kw_list)):
                r_kwarg[kw_list[index]] = r.json[kw_list[index]]
    else:
        if not kw_list:
            r_arg = is_form(r.form)
        else:
            for index in range(len(kw_list)):
                r_kwarg[kw_list[index]] = r.form[kw_list[index]]
    return f(*r_arg, **r_kwarg)


def is_json(arg_list):
    return arg_list.values()


def is_form(arg_list):
    return arg_list.to_dict().values()


# 路径绑定装饰器
# 默认服务器从boot获取
def request_mapping(path='', methods=[], app=None):
    def method(f):
        try:
            # 自动获取蓝图实例并进行http协议绑定
            current_package = __import__(str(getattr(f, "__module__")), fromlist=True)
            try:
                package_app = getattr(current_package, '__app') \
                    if hasattr(current_package, '__app') \
                    else current_package.app \
                    if hasattr(current_package, 'app') \
                    else app \
                    if hasattr(app, 'route') \
                    else __import__('dophon').blue_print(f"_annotation_auto_reg{getattr(current_package, '__name__')}",
                                                         getattr(current_package, '__name__'))
            except Exception as e:
                logger.warn(f'{e}')
                package_app = __import__('dophon.boot', fromlist=True).app
            # 回设内部蓝图参数
            # print(package_app)
            setattr(current_package, '__app', package_app)
            result = package_app.route(path, methods=methods)(f)
        except Exception as e:
            logger.error(f'{getattr(f, "__module__")}参数配置缺失,请检查({path},{methods},{package_app})')

        def m_args(*args, **kwargs):
            return result(*args, **kwargs)

        return m_args

    return method


# get方法缩写
def get_route(path=''):
    def method(f):
        result = request_mapping(path, ['get'])(f)

        def m_args(*args, **kwargs):
            return result(*args, **kwargs)

        return m_args

    return method


# post方法缩写
def post_route(path=''):
    def method(f):
        result = request_mapping(path, ['post'])(f)

        def args(*args, **kwargs):
            return result(*args, **kwargs)

        return args

    return method


# get方法缩写
def get(f, *args, **kwargs):
    path = f'{"/" if re.match("^[a-zA-Z0-9]+", getattr(f, "__name__")) else ""}{re.sub("[^a-zA-Z0-9]", "/",getattr(f, "__name__"))}'
    result = request_mapping(re.sub('\s+', '', path), ['get'])(f)


    def method():
        def args(*args, **kwargs):
            return result(*args, **kwargs)

        return args


    return method


# post方法缩写
def post(f, *args, **kwargs):
    path = f'{"/" if re.match("^[a-zA-Z0-9]+", getattr(f, "__name__")) else ""}{re.sub("[^a-zA-Z0-9]", "/",getattr(f, "__name__"))}'
    result = request_mapping(re.sub('\s+', '', path), ['post'])(f)


    def method():
        def args(*args, **kwargs):
            return result(*args, **kwargs)

        return args


    return method



# coding: utf-8
from dophon.annotation import AutoWired
from dophon.annotation import res, req
from dophon.annotation.AutoWired import *

"""
注解集合(部分)
author:CallMeE
date:2018-06-01


"""
__all__ = [
    'ResponseBody',
    'ResponseTemplate',
    'AutoParam',
    'FullParam',
    'RequestMapping',
    'GetRoute',
    'PostRoute',
    'Get',
    'Post',
    'AutoWired',
    'AsResBody',
    'AsResTemp',
    'AsArgs',
    'AsJson',
    'AsFile',
    'BeanConfig',
    'bean',
    'Bean'
]

AutoWired = AutoWired

ResponseBody = AsResBody = res.response_body

ResponseTemplate = AsResTemp = res.response_template

AutoParam = AsArgs = req.auto_param

FullParam = AsJson = req.full_param

FileParam = AsFile = req.file_param

BeanConfig = AutoWired.BeanConfig

bean = AutoWired.bean

Bean = AutoWired.Bean

RequestMapping = req.request_mapping

GetRoute = req.get_route

PostRoute = req.post_route

Get = req.get

Post = req.post



# coding: utf-8
from gevent import monkey

monkey.patch_all()
from multiprocessing import Process, freeze_support
import time, socket, random
from flask import request, make_response
from urllib3 import PoolManager
from dophon import properties
from dophon_logger import *

logger = get_logger(DOPHON)

logger.inject_logger(globals())

ports = []  # 记录监听端口

proxy_clusters = {}

pool = PoolManager()


def main_freeze():
    freeze_support()


def redirect_request():
    logger.info('touch path: %s [success]' % (request.path))
    res = pool.request(request.method, '127.0.0.1:' + str(random.choice(ports)) + request.path,
                       fields=request.json if request.is_json else request.form)
    return make_response(res.data)


def outer_entity(boot):
    # 重写路由信息(修改为重定向路径)
    boot.get_app().before_request(redirect_request)
    boot.run_app()


def run_clusters(clusters: int, outer_port: bool = False, start_port: int = 8800):
    """
    运行集群式服务器
    :param clusters: 集群个数
    :param outer_port: 是否开启外部端口映射(映射端口为用户配置文件中配置的端口)
    :param start_port: 集群起始监听端口
    :return:
    """
    from dophon import boot
    for i in range(clusters):
        current_port = start_port + i
        create_cluster_cell(boot=boot, port=current_port)
        ports.append(current_port)
    while len(ports) != clusters:
        time.sleep(5)

    logger.info('启动检测端口监听')
    for port in ports:
        if check_socket(int(port)):
            continue
    logger.info('集群端口: %s ' % ports)
    if outer_port:
        logger.info('启动外部端口监听[%s]' % (properties.port))
        outer_entity(boot)


def create_cluster_cell(boot, port):
    # http协议
    proc = Process(target=boot.run_app, kwargs={
        'host': '127.0.0.1',
        'port': port
    })
    proc.start()


def check_socket(port: int):
    s = socket.socket()
    flag = True
    while flag:
        try:
            ex_code = s.connect_ex(('127.0.0.1', port))
            flag = False
            return int(ex_code) == 0
        except Exception as e:
            time.sleep(3)
            continue



# coding: utf-8
import traceback
import functools
import os
import sys
import re
from datetime import datetime
from threading import *
import json

"""
初始化协程模块(必须,不然导致系统死锁)
"""
from gevent import monkey

monkey.patch_all()

from dophon import pre_boot
from dophon_logger import *

logger = get_logger(DOPHON)

logger.inject_logger(globals())

pre_boot.check_modules()

from flask import Flask, request, abort, jsonify
from dophon import properties, blue_print
from dophon import tools
from dophon.tools import gc

try:
    mysql = __import__('dophon.db.mysql')
except:
    mysql = None


def load_banner():
    """
    加载header-banner文件
    :return:
    """
    file_root = properties.project_root
    file_path = file_root + os.path.sep + 'header.txt'
    if os.path.exists(file_path):
        with open(file_path, encoding='utf8') as banner:
            for line in banner.readlines():
                sys.stdout.write(line)
    else:
        tools.show_banner()
    sys.stdout.flush()


def load_footer():
    """
    加载footer-banner文件
    :return:
    """
    file_root = properties.project_root
    file_path = file_root + os.path.sep + 'footer.txt'
    if os.path.exists(file_path):
        with open(file_path, encoding='utf8') as banner:
            for line in banner.readlines():
                sys.stdout.write(line)
    sys.stdout.flush()


error_info = properties.error_info


def boot_init():
    """
    初始化启动
    :return:
    """
    global app_name, app, ip_count, ipcount_lock, ip_refuse_list
    load_banner()
    app_name = properties.service_name if hasattr(properties, 'service_name') else __name__
    # 定义WEB容器(同时防止json以ascii解码返回)
    app = Flask(app_name)
    app.config['JSON_AS_ASCII'] = False

    # ip计数缓存
    ip_count = {}
    ipcount_lock = Lock()
    if os.path.exists(os.getcwd() + '/ip_count'):
        with open('ip_count', 'r') as ip_count_file:
            file_text = ip_count_file.read()
            if file_text:
                ip_count = eval(file_text)
    else:
        with open('ip_count', 'w') as ip_count_file:
            json.dump({}, ip_count_file)
    # IP黑名单
    ip_refuse_list = {}


boot_init()


def before_request():
    """
    定义拦截器
    :return:
    """
    if properties.ip_count:
        ipcount_lock.locked()
        ip = request.remote_addr
        now_timestemp = datetime.now().timestamp()
        # 检测是否为黑名单
        if ip in ip_refuse_list:
            # 默认禁用一分钟
            if (int(now_timestemp) - int(ip_refuse_list[ip])) > 60:
                # 可以清除白名单
                ip_refuse_list.pop(ip)
                # 清除访问记录
                ip_count[ip]['req_timestemp'] = [now_timestemp]
            else:
                # 未到解禁时间
                return abort(400)
        if ip in ip_count:
            ip_item = ip_count[ip]
            ip_item['req_count'] += 1
            req_timestemp = ip_item['req_timestemp']
            """
            判断逻辑:
                当前请求时间是最近的,连续的1秒内
            """
            if (int(now_timestemp) - int(req_timestemp[0])) > 1 \
                    and \
                    (int(now_timestemp) - int(req_timestemp[len(req_timestemp) - 1])) < 1:
                # 检测3秒内请求数
                if len(req_timestemp) > 50:
                    # 默认三秒内请求不超过300
                    # 超出策略则添加到黑名单
                    ip_refuse_list[ip] = now_timestemp
                else:
                    # 不超出策略则清空请求时间戳缓存
                    ip_item['req_timestemp'] = [now_timestemp]
            else:
                ip_item['req_timestemp'].append(now_timestemp)
        else:
            ip_item = {
                'req_count': 1,
                'req_timestemp': [now_timestemp]
            }
        Thread(target=persist_ip_count).start()
        ip_count[ip] = ip_item


def persist_ip_count():
    """
    持久化ip统计到文件
    :return:
    """
    # 持久化ip统计
    with open('ip_count', 'w') as ip_count_file:
        json.dump(ip_count, ip_count_file)
    if ipcount_lock.locked():
        ipcount_lock.release()


blueprint_init_queue = {}  # 蓝图初始化方法缓存(用于初始化后启动)


# 处理各模块中的自动注入以及组装各路由
# dir_path中为路由模块路径,例如需要引入的路由都在routes文件夹中,则传入参数'/routes'
def map_apps(dir_path):
    path = os.getcwd() + dir_path
    if not os.path.exists(path):
        logger.error('路由文件夹不存在,创建路由文件夹')
        os.mkdir(path)
    f_list = os.listdir(path)
    logger.info(f'路由文件夹: {dir_path}')
    while f_list:
        try:
            file = f_list.pop(0)
            if re.match('__.*__', file):
                continue
            i = os.path.join(path, file)
            if os.path.isdir(i):
                logger.info(f'加载路由模块: {file}')
                continue
            file_name = re.sub('\.py', '', file)
            f_model = __import__(re.sub('/', '', dir_path) + '.' + file_name, fromlist=True)
            # 自动装配蓝图实例并自动配置部分参数,免除繁琐配置以及精简代码
            package_app = getattr(f_model, '__app') \
                if hasattr(f_model, '__app') \
                else f_model.app \
                if hasattr(f_model, 'app') \
                else blue_print(f"_boot_auto_reg_{file_name}", getattr(f_model, '__name__'))
            filter_method = package_app.before_request(before_request)
            # 若需统计请求,装配请求统计方法
            if hasattr(properties, 'ip_count') and getattr(properties, 'ip_count'):
                setattr(f_model, 'before_request', filter_method)
            # 若存在初始化执行方法,执行该方法
            if hasattr(f_model, 'blueprint_init'):
                init_fun = getattr(f_model, 'blueprint_init')
                # 判断是否方法
                if callable(init_fun):
                    blueprint_init_queue[f_model] = before_bp_init_fun(init_fun)
            app.register_blueprint(package_app)
        except Exception as e:
            raise e
            pass

    # for item in get_app().url_map.iter_rules():
    #     print(item)
    # print(get_app().blueprints)
    # 注册路径列表入口
    get_app().route('/rule/<view>')(lambda view: f"""
<!-- import Vue.js -->
<script src="//vuejs.org/js/vue.min.js"></script>
<!-- import stylesheet -->
<link rel="stylesheet" href="//unpkg.com/iview/dist/styles/iview.css">
<!-- import iView -->
<script src="//unpkg.com/iview/dist/iview.min.js"></script>
<style>
.ivu-card""""{width: 100%; display: inline-block; margin-top: 10px; margin-bottom: 10px;}"f"""
</style>
<div style="display;flex;column-count:5;">
<div class="ivu-card ivu-card-bordered"><div class="ivu-card-body">
    {'</div></div><div class="ivu-card ivu-card-bordered"><div class="ivu-card-body">'.join(
        [
            str(item)
            for item in get_app().url_map.iter_rules()

        ]
    )}
</div></div>
</div>
    """ if view == 'map' else jsonify([
        str(item)
        for item in get_app().url_map.iter_rules()

    ]) if view == 'json' else abort(404)
                                 )


def before_bp_init_fun(f):
    """
    预留蓝图初始化装饰器
    :param f:
    :return:
    """

    # print(f)

    def fields(*args, **kwargs):
        # print('args: ', args, 'kwargs:', kwargs)
        f(*args, **kwargs)

    return fields


def get_app() -> Flask:
    return app


def free_source():
    def method(f):
        @functools.wraps(f)
        def args(*arg, **kwarg):
            logger.info('启动服务器')
            logger.info('路由初始化')
            for path in properties.blueprint_path:
                map_apps(path)
            load_footer()
            # 执行蓝图初始化方法
            for blueprint_module, blue_print_init_method in blueprint_init_queue.items():
                try:
                    blue_print_init_method()
                except Exception as e:
                    logger.error(f'蓝图"{blueprint_module}"初始化失败,信息: {e}')
            # for rule in get_app().url_map.iter_rules():
            #     print(str(rule))
            # print(rule.get_rules())
            # for name in dir(rule):
            #     print(f'{name}---{getattr(rule,name)}')
            # print(help(rule))
            # break
            f(*arg, **kwarg)
            """
            释放所有资源
            :return:
            """
            logger.info('服务器关闭')
            logger.info('释放资源')
            if mysql:
                mysql.free_pool()
            logger.info('释放连接池')
            sys.exit()
            # logger.info('再次按下Ctrl+C退出')

        return args

    return method


def fix_static(
        fix_target: Flask = app,
        static_floder: str = 'static',
        enhance_power: bool = False
):
    """
    修正静态文件路径
    :return:
    """
    if hasattr(fix_target, 'static_folder'):
        if hasattr(properties, 'project_root'):
            root_path = properties.project_root
        else:
            root_path = os.getcwd()
        static_floder_path = root_path + '/' + static_floder
        fix_target.static_folder = static_floder_path
        if os.path.exists(static_floder_path) and enhance_power:
            enhance_static_route(static_floder_path)
    else:
        raise Exception('错误的修复对象')


def enhance_static_route(static_floder_path: str):
    """
    增强静态文件路由能力
    :param static_floder_path:
    :param serlize:
    :return:
    """
    framework_static_route_path = f'{properties.project_root}{properties.blueprint_path[0]}/FrameworkStaticRoute.py'
    # if not os.path.exists(framework_static_route_path):
    if True:
        import types
        import uuid
        # 定义静态资源获取路径
        logger.info('增强静态文件路由')
        with open(framework_static_route_path, 'wb') as fsroute:
            # 写入预设信息
            fsroute.write(bytes(
                f"# -*- coding: utf-8 -*-\n"
                f"from dophon import blue_print\n"
                f"from dophon.annotation import *\n"
                f"from flask import url_for\n"
                f"from flask import render_template\n"
                f"from flask import send_from_directory\n"
                f"app = blue_print('FrameworkStaticRoute', __name__,static_folder='{static_floder_path}')\n",
                encoding='utf-8'))
            for root, dir_path, file in os.walk(static_floder_path):
                # 消除转义字符(win)
                root = re.sub('\\\\', '/', root)
                static_floder_path = re.sub('\\\\', '/', static_floder_path)
                # 静态目录下的目录名
                root_dir_path = re.sub('\\\\', '/', re.sub(static_floder_path, "", root))
                # 解析静态资源路径
                route_name = re.sub('[^(1-9a-zA-Z_)]', '',
                                    f'{re.sub(static_floder_path, "", root)}_{root_dir_path}_{uuid.uuid1()}')
                route_path = f'{root_dir_path}/<file_name>'
                static_url_method_code_body = \
                    f"@RequestMapping('{route_path}',['get','post'])\ndef {route_name}(file_name):\n\t" \
                        f"return render_template(f'{root_dir_path}/" \
                        "{file_name}" \
                        f"') if file_name.endswith('.html') " \
                        f"else send_from_directory('{static_floder_path}{root_dir_path}',f" \
                        "'{file_name}'" \
                        ",as_attachment=True)\n"
                fsroute.write(bytes(
                    static_url_method_code_body,
                    encoding='utf-8'))
        logger.info("增强静态文件夹完毕")


def fix_template(
        fix_target: Flask = app,
        template_folder: str = 'templates'
):
    """
    修正页面模板文件路径
    :return:
    """
    if hasattr(fix_target, 'template_folder'):
        if hasattr(properties, 'project_root'):
            root_path = properties.project_root
        else:
            root_path = os.getcwd()
        fix_target.template_folder = root_path + '/' + template_folder
    else:
        raise Exception('错误的修复对象')


@free_source()
def run_app(host=properties.host, port=properties.port):
    logger.info(f'监听地址: {host} : {port}')
    if properties.server_gevented:
        from gevent.pywsgi import WSGIServer
        WSGIServer((host, port), app).serve_forever()
    else:
        if properties.server_threaded:
            # 开启多线程处理
            app.run(host=host, port=port, threaded=properties.server_threaded)
        elif tools.is_not_windows() and properties.server_processes > 1:
            # 开启多进程处理
            print('开启多进程', properties.server_processes)
            app.run(host=host, port=port, threaded=False, processes=properties.server_processes)
        else:
            app.run(host=host, port=port)


@free_source()
def run_app_ssl(host=properties.host, port=properties.port, ssl_context=properties.ssl_context):
    logger.info(f'监听地址: {host} : {port}')
    if properties.server_gevented:
        from gevent.pywsgi import WSGIServer
        ssl_args = {
            'certfile': ssl_context[0],
            'keyfile': ssl_context[1],
        }
        WSGIServer((host, port), app, **ssl_args).serve_forever()
    else:
        if properties.server_threaded:
            # 开启多线程处理
            app.run(host=host, port=port, ssl_context=ssl_context, threaded=properties.server_threaded)
        elif tools.is_not_windows() and properties.server_processes > 1:
            # 开启多进程处理
            app.run(host=host, port=port, ssl_context=ssl_context, threaded=False,
                    processes=properties.server_processes)
        else:
            app.run(host=host, port=port, ssl_context=ssl_context)


def bootstrap_app():
    """
    bootstrap样式页面初始化
    :return:
    """
    global app
    b = __import__('flask_bootstrap')
    b.Bootstrap(app)


from dophon.annotation import *


@RequestMapping('/framework/ip/count', ['get', 'post'])
@ResponseBody()
def view_ip_count():
    """
    此为公共请求,不计入ip统计
    :return:
    """
    return ip_count


@RequestMapping('/framework/ip/refuse', ['get', 'post'])
@ResponseBody()
def view_ip_refuse():
    """
    此为公共请求,不计入ip统计
    :return:
    """
    return ip_refuse_list


GC_INFO = False  # gc信息开关


@RequestMapping('/framework/gc/show', ['get'])
@ResponseBody()
def show_gc_info():
    # return gc.show_gc_leak(print)
    return gc.show_gc_leak(logger.info) if GC_INFO else {}


from dophon.tools.framework_const import error_info_type, self_result


@app.errorhandler(500)
def handle_500(e):
    """
    处理业务代码异常页面
    :param e:
    :return:
    """
    exc_type, exc_value, exc_tb = sys.exc_info()
    tc = traceback.format_tb(exc_tb)
    if properties.debug_trace:
        trace_detail = '<h4>Details</h4>'
        trace_detail += '<table style="width:100%;" cellspacing="0" cellpadding="4">'
        for tc_item in tc:
            if tc_item.find(os.getcwd()) < 0:
                continue
            trace_detail += '<tr>'
            tc_item_info_list = tc_item.split(',', 2)
            # for tc_item_info_list_item in tc_item_info_list:
            #     trace_detail += '<td>' + tc_item_info_list_item + '</td>'
            trace_detail += '<td style="border: 1px solid gray;">' + re.sub('(^.+(\\\|/))|"', '',
                                                                            tc_item_info_list[0]) + '</td>'
            trace_detail += '<td style="border: 1px solid gray;">' + tc_item_info_list[1] + '</td>'
            trace_code_detail = re.sub('^\s+', '', tc_item_info_list[2]).split(' ', 1)
            trace_detail += '<td style="border: 1px solid gray;">' + ' => '.join([
                trace_code_detail[0],
                re.sub('\s+', ' => ', trace_code_detail[1], 1),
            ]) + '</td>'
            trace_detail += '<tr>'
        trace_detail += '</table>'
    else:
        trace_detail = ''
    if error_info == error_info_type.HTML:
        return '<h1>Wrong!!</h1>' + \
               '<h2>error info:' + str(e) + '</h2>' + \
               '<h3>please contact coder or direct to ' \
               '<a href="https://dophon.blog">dophon</a> and leave your question</h3>' + \
               trace_detail, 500
    elif error_info == error_info_type.JSON:
        return self_result.JsonResult(500, tc, """
        please contact coder or direct to dophon website and leave your question
        """).as_res()
    elif error_info == error_info_type.XML:
        return self_result.XmlResult(500, tc, """
        please contact coder or direct to dophon website and leave your question
        """).as_res()


@app.errorhandler(404)
def handle_404(e):
    """
    处理路径匹配异常
    :return:
    """
    global error_info
    if error_info == error_info_type.HTML:
        return '<h1>Wrong!!</h1>' + \
               '<h2>error info:' + str(e) + '</h2>' + \
               '<h3>please contact coder or direct to ' \
               '<a href="https://dophon.blog">dophon</a> and leave your question</h3>' + \
               'request path:' + request.path, 404
    elif error_info == error_info_type.JSON:
        return self_result.JsonResult(404, request.path, 'please check your path').as_res()
    elif error_info == error_info_type.XML:
        return self_result.XmlResult(404, request.path, 'please check your path').as_res()


@app.errorhandler(405)
def handle_405(e):
    """
    处理请求方法异常
    :return:
    """
    if error_info == error_info_type.HTML:
        return '<h1>Wrong!!</h1>' + \
               '<h2>error info:' + str(e) + '</h2>' + \
               '<h3>please contact coder or direct to ' \
               '<a href="https://dophon.blog">dophon</a> and leave your question</h3>' + \
               'request method:' + request.method, 405
    elif error_info == error_info_type.JSON:
        return self_result.JsonResult(405, {request.path, request.method}, 'please check your request method').as_res()
    elif error_info == error_info_type.XML:
        return self_result.XmlResult(405, {request.path, request.method}, 'please check your request method').as_res()


@app.errorhandler(400)
def handle_400(e):
    """
    处理异常请求
    :return:
    """
    if error_info == error_info_type.HTML:
        return ('<h1>Wrong!!</h1>' + \
                '<h2>error info:' + str(e) + '</h2>' + \
                '<h3>please contact coder or direct to '
                '<a href="https://dophon.blog">dophon</a> and leave your question</h3>' + \
                'request form:' + request.form + \
                'request body:' + request.json if request.is_json else ''), 400
    elif error_info == error_info_type.JSON:
        return self_result.JsonResult(400, request.json if request.is_json else '',
                                      'please check your request data').as_res()
    elif error_info == error_info_type.XML:
        return self_result.XmlResult(400, request.json if request.is_json else '',
                                     'please check your request data').as_res()



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['dophon',
 'dophon.annotation',
 'dophon.annotation.req',
 'dophon.annotation.res',
 'dophon.tools',
 'dophon.tools.framework_const',
 'dophon.tools.gc']

package_data = \
{'': ['*']}

install_requires = \
['dophon-logger',
 'dophon-manager',
 'dophon-properties',
 'flask',
 'gevent',
 'pyyaml>=3.13,<4.0',
 'schedule',
 'urllib3']

setup_kwargs = {
    'name': 'dophon',
    'version': '1.2.13.post2',
    'description': 'dophon web framework like springboot',
    'long_description': None,
    'author': 'CallMeE',
    'author_email': 'ealohu@163.com',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
